
#ifdef __cplusplus
extern "C" {            /* Assume C declarations for C++ */
#endif	/* __cplusplus */

void* FAR PASCAL _export CntrDLLContainerObjectNew();
void FAR PASCAL _export CntrDLLContainerObjectDelete(void* pData);
BOOL FAR PASCAL _export CntrDLLContainerObjectActivate(void* pItem, void _huge* pData, long nSize);

void* FAR PASCAL _export CntrDLLInsertObjectNew();
void FAR PASCAL _export CntrDLLInsertObjectDelete(void *pData);
HGLOBAL FAR PASCAL _export CntrDLLInsertObjectGetHGlobal(void* pData);

#ifdef __cplusplus
}                       /* End of extern "C" { */
#endif	/* __cplusplus */
